@@ -1,6 +1,6 @@
 /* 
    HTTP Request Handling
-   Copyright (C) 1999-2006, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2024, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -44,27 +44,24 @@ typedef struct ne_request_s ne_request;
 
 /***** Request Handling *****/
 
-/* Create a request in session 'sess', with given method and path.
- * 'path' must conform to the 'abs_path' grammar in RFC2396, with an
- * optional "? query" part, and MUST be URI-escaped by the caller. */
-ne_request *ne_request_create(ne_session *sess,
-			      const char *method, const char *path);
+/* Create a request in session 'sess', with given method and target.
+ * 'target' is used to form the request-target (per RFC 9112ẞ3.2), and
+ * may be an absolute-path (with optional query-string), an
+ * absolute-URI, or an asterisk. */
+ne_request *ne_request_create(ne_session *sess, const char *method,
+                              const char *target)
+    ne_attribute((nonnull));
 
 /* The request body will be taken from 'size' bytes of 'buffer'. */
 void ne_set_request_body_buffer(ne_request *req, const char *buffer,
-				size_t size);
+                                size_t size)
+    ne_attribute((nonnull));
 
 /* The request body will be taken from 'length' bytes read from the
  * file descriptor 'fd', starting from file offset 'offset'. */
 void ne_set_request_body_fd(ne_request *req, int fd,
-                            off_t offset, off_t length);
-
-#ifdef NE_LFS
-/* Alternate version of ne_set_request_body_fd taking off64_t 
- * offset type for systems supporting _LARGEFILE64_SOURCE. */
-void ne_set_request_body_fd64(ne_request *req, int fd,
-                              off64_t offset, off64_t length);
-#endif
+                            ne_off_t offset, ne_off_t length)
+    ne_attribute((nonnull));
 
 /* "Pull"-based request body provider: a callback which is invoked to
  * provide blocks of request body on demand.
@@ -79,22 +76,22 @@ void ne_set_request_body_fd64(ne_request *req, int fd,
  *        <0           : error, abort request; session error string must be set.
  *         0           : ignore 'buffer' contents, end of body.
  *     0 < x <= buflen : buffer contains x bytes of body data.  */
-typedef ssize_t (*ne_provide_body)(void *userdata, 
-				   char *buffer, size_t buflen);
+typedef ssize_t (*ne_provide_body)(void *userdata, char *buffer, size_t buflen);
 
 /* Install a callback which is invoked as needed to provide the
  * request body, a block at a time.  The total size of the request
  * body is 'length'; the callback must ensure that it returns no more
- * than 'length' bytes in total. */
-void ne_set_request_body_provider(ne_request *req, off_t length,
-				  ne_provide_body provider, void *userdata);
-
-#ifdef NE_LFS
-/* Duplicate version of ne_set_request_body_provider, taking an off64_t
- * offset. */
-void ne_set_request_body_provider64(ne_request *req, off64_t length,
-                                    ne_provide_body provider, void *userdata);
-#endif
+ * than 'length' bytes in total.  If 'length' is set to -1, then the
+ * total size of the request is unknown by the caller and chunked
+ * transfer will be used.
+ *
+ * The caller MUST determine that the server can accept chunked
+ * encoding (i.e. advertises HTTP/1.1 support) before using chunked
+ * encoding. This can be done by testing that ne_version_pre_http11()
+ * returns zero after performing an OPTIONS or HEAD request. */
+void ne_set_request_body_provider(ne_request *req, ne_off_t length,
+                                  ne_provide_body provider, void *userdata)
+    ne_attribute((nonnull (1, 3)));
 
 /* Handling response bodies; two callbacks must be provided:
  *
@@ -137,13 +134,15 @@ typedef int (*ne_block_reader)(void *userdata, const char *buf, size_t len);
  * been read, the callback will be called with a 'len' argument of
  * zero.  */
 void ne_add_response_body_reader(ne_request *req, ne_accept_response accpt,
-				 ne_block_reader reader, void *userdata);
+				 ne_block_reader reader, void *userdata)
+    ne_attribute((nonnull (1, 2, 3)));
 
 /* Retrieve the value of the response header field with given name;
  * returns NULL if no response header with given name was found.  The
  * return value is valid only until the next call to either
  * ne_request_destroy or ne_begin_request for this request. */
-const char *ne_get_response_header(ne_request *req, const char *name);
+const char *ne_get_response_header(ne_request *req, const char *name)
+    ne_attribute((nonnull));
 
 /* Iterator interface for response headers: if passed a NULL cursor,
  * returns the first header; if passed a non-NULL cursor pointer,
@@ -168,6 +167,22 @@ void ne_print_request_header(ne_request *req, const char *name,
 			     const char *format, ...) 
     ne_attribute((format(printf, 3, 4)));
 
+/* Returns the request target URI as a malloc-allocated ne_uri object,
+ * or NULL on error if the request target cannot be determined. The
+ * session error string is not changed on error. */
+const ne_uri *ne_get_request_target(ne_request *req)
+    ne_attribute((nonnull));
+
+/* If the response includes a Location header, this function parses
+ * and resolves the URI-reference relative to the request target.  If
+ * a fragment ("#fragment") is used for the request target, it can be
+ * passed as an argument to allow relative resolution. Returns a
+ * malloc-allocated ne_uri object, or NULL if the URI in the Location
+ * header could not be parsed, or the Location header was not
+ * present. */
+ne_uri *ne_get_response_location(ne_request *req, const char *fragment)
+    ne_attribute((nonnull (1)));
+
 /* ne_request_dispatch: Sends the given request, and reads the
  * response.  Returns:
  *  - NE_OK if the request was sent and response read successfully
@@ -178,7 +193,7 @@ void ne_print_request_header(ne_request *req, const char *name,
  * On any error, the session error string is set.  On success or
  * authentication error, the actual response-status can be retrieved using
  * ne_get_status(). */
-int ne_request_dispatch(ne_request *req);
+int ne_request_dispatch(ne_request *req) ne_attribute((nonnull (1)));
 
 /* Returns a pointer to the response status information for the given
  * request; pointer is valid until request object is destroyed. */
@@ -232,6 +247,10 @@ typedef enum ne_request_flag_e {
     NE_REQFLAG_IDEMPOTENT, /* disable this flag if the request uses a
                             * non-idempotent method such as POST. */
 
+    NE_REQFLAG_1XXTIMEOUT, /* disable this flag to apply no overall
+                             * timeout when reading interim
+                             * responses. */
+
     NE_REQFLAG_LAST /* enum sentinel value */
 } ne_request_flag;
 
@@ -242,25 +261,44 @@ void ne_set_request_flag(ne_request *req, ne_request_flag flag, int value);
  * flag is not supported. */
 int ne_get_request_flag(ne_request *req, ne_request_flag flag);
 
+/* Callback to handle an interim (1xx) response. The status-code of
+ * the response is passed as 'status'; interim response headers can be
+ * accessed via ne_get_response_header. */
+typedef void (*ne_interim_response_fn)(void *userdata, ne_request *req,
+                                       const ne_status *status);
+
+/* Add a interim response callback handler for the request. */
+void ne_add_interim_handler(ne_request *req, ne_interim_response_fn fn,
+                            void *userdata);
+
 /**** Request hooks handling *****/
 
-typedef void (*ne_free_hooks)(void *cookie);
-
-/* Hook called when a request is created; passed the request method,
- * and the string used as the Request-URI (note that this may be a
- * absolute URI if a proxy is in use, an absolute path, a "*",
- * etc). */
+/* Hook called when a request is created; passed the method and
+ * request-target as used in the request-line (RFC 9112§3.2).  The
+ * create_request hook is called exactly once per request. */
 typedef void (*ne_create_request_fn)(ne_request *req, void *userdata,
-				     const char *method, const char *requri);
+				     const char *method, const char *target);
 void ne_hook_create_request(ne_session *sess, 
 			    ne_create_request_fn fn, void *userdata);
 
 /* Hook called before the request is sent.  'header' is the raw HTTP
- * header before the trailing CRLF is added: add in more here. */
+ * header before the trailing CRLF is added; more headers can be added
+ * here.  A pre_send hook may be called >1 time per request if the
+ * request is retried due to a post_send hook returning NE_RETRY. */
 typedef void (*ne_pre_send_fn)(ne_request *req, void *userdata, 
 			       ne_buffer *header);
 void ne_hook_pre_send(ne_session *sess, ne_pre_send_fn fn, void *userdata);
 
+/* Hook called directly after the response headers have been read, but
+ * before the resposnse body has been read.  'status' is the response
+ * status-code.  A post_header hook may be called >1 time per request
+ * if the request is retried due to a post_send hook returning
+ * NE_RETRY. */
+typedef void (*ne_post_headers_fn)(ne_request *req, void *userdata,
+                                   const ne_status *status);
+void ne_hook_post_headers(ne_session *sess, 
+                          ne_post_headers_fn fn, void *userdata);
+
 /* Hook called after the request is dispatched (request sent, and
  * the entire response read).  If an error occurred reading the response,
  * this hook will not run.  May return:
@@ -280,10 +318,15 @@ void ne_hook_destroy_request(ne_session *sess,
 			     ne_destroy_req_fn fn, void *userdata);
 
 typedef void (*ne_destroy_sess_fn)(void *userdata);
-/* Hook called when the session is destroyed. */
+/* Hook called when the session is about to be destroyed. */
 void ne_hook_destroy_session(ne_session *sess,
 			     ne_destroy_sess_fn fn, void *userdata);
 
+typedef void (*ne_close_conn_fn)(void *userdata);
+/* Hook called when the connection is closed; note that this hook
+ * may be called *AFTER* the destroy_session hook. */
+void ne_hook_close_conn(ne_session *sess, ne_close_conn_fn fn, void *userdata);
+
 /* The ne_unhook_* functions remove a hook registered with the given
  * session.  If a hook is found which was registered with a given
  * function 'fn', and userdata pointer 'userdata', then it will be
@@ -294,11 +337,14 @@ void ne_hook_destroy_session(ne_session *sess,
 void ne_unhook_create_request(ne_session *sess, 
                               ne_create_request_fn fn, void *userdata);
 void ne_unhook_pre_send(ne_session *sess, ne_pre_send_fn fn, void *userdata);
+void ne_unhook_post_headers(ne_session *sess, ne_post_headers_fn fn, void *userdata);
 void ne_unhook_post_send(ne_session *sess, ne_post_send_fn fn, void *userdata);
 void ne_unhook_destroy_request(ne_session *sess,
                                ne_destroy_req_fn fn, void *userdata);
 void ne_unhook_destroy_session(ne_session *sess,
                                ne_destroy_sess_fn fn, void *userdata);
+void ne_unhook_close_conn(ne_session *sess, 
+                          ne_close_conn_fn fn, void *userdata);
 
 /* Store an opaque context for the request, 'priv' is returned by a
  * call to ne_request_get_private with the same ID. */
